import numpy as np

def damerau_levenshtein_distance(a, b):
    INF = len(a) + len(b)
    matrix  = [[INF for n in xrange(len(b) + 2)]]
    matrix += [[INF] + range(len(b) + 1)]
    matrix += [[INF, m] + [0] * len(b) for m in xrange(1, len(a) + 1)]

    last_row = {}

    for row in xrange(1, len(a) + 1):
        ch_a = a[row-1]

        last_match_col = 0

        for col in xrange(1, len(b) + 1):
            ch_b = b[col-1]

            last_matching_row = last_row.get(ch_b, 0)
            cost = 0 if ch_a == ch_b else 1

            matrix[row+1][col+1] = min(
                matrix[row][col] + cost, # Substitution
                matrix[row+1][col] + 1,  # Addition
                matrix[row][col+1] + 1,  # Deletion

                # Transposition
                matrix[last_matching_row][last_match_col]
                    + (row - last_matching_row - 1) + 1
                    + (col - last_match_col - 1))

            if cost == 0:
                last_match_col = col

        last_row[ch_a] = row
    
	return matrix[-1][-1]

def levenshtein_distance(source, target):
    if len(source) < len(target):
        return levenshtein(target, source)

    if len(target) == 0:
        return len(source)

    source = np.array(tuple(source))
    target = np.array(tuple(target))

    previous_row = np.arange(target.size + 1)
    for s in source:
        current_row = previous_row + 1

        current_row[1:] = np.minimum(
                current_row[1:],
                np.add(previous_row[:-1], target != s))

        current_row[1:] = np.minimum(
                current_row[1:],
                current_row[0:-1] + 1)

        previous_row = current_row

    return previous_row[-1]



def tied_rank(x):
    sorted_x = sorted(zip(x,range(len(x))))
    r = [0 for k in x]
    cur_val = sorted_x[0][0]
    last_rank = 0
    for i in range(len(sorted_x)):
        if cur_val != sorted_x[i][0]:
            cur_val = sorted_x[i][0]
            for j in range(last_rank, i): 
                r[sorted_x[j][1]] = float(last_rank+1+i)/2.0
            last_rank = i
        if i==len(sorted_x)-1:
            for j in range(last_rank, i+1): 
                r[sorted_x[j][1]] = float(last_rank+i+2)/2.0
    return r

def auc(actual, posterior):
    r = tied_rank(posterior)
    num_positive = len([0 for x in actual if x==1])
    num_negative = len(actual)-num_positive
    sum_positive = sum([r[i] for i in range(len(r)) if actual[i]==1])
    auc = ((sum_positive - num_positive*(num_positive+1)/2.0) /
           (num_negative*num_positive))
    return auc



from distutils.core import setup
requirements = ["numpy", "pandas"]

setup(
  name = 'mltk',
  packages = ['mltk', 'mltk.metrics'], # this must be the same as the name above
  version = '0.0.4',
  description = 'Machine learning toolkit for python',
  author = 'Manan Shah',
  author_email = 'manan.shah.777@gmail.com',
  url = 'https://github.com/mananshah99/mltk', # use the URL to the github repo
  keywords = ['ml', 'metrics', 'toolkit'], # arbitrary keywords
  install_requires = requirements,
  classifiers = [],
)



import numpy as np

class ConfusionMatrix:

    def __init__(self, n_classes):
        self.n_classes = n_classes
        self.mat = np.zeros((n_classes,n_classes), dtype='int')
        self.lista = np.array(())
        self.listb = np.array(())

    def __str__(self):
        return np.array_str(self.mat)

    def add(self, y_true, y_pred):
        assert len(y_true) == len(y_pred)
        assert max(y_true) < self.n_classes
        assert max(y_pred) < self.n_classes

        self.lista = np.append(self.lista, np.asarray(y_true)).flatten()
        self.listb = np.append(self.listb, np.asarray(y_pred)).flatten()

        for i in range(len(y_true)):
                self.mat[y_true[i],y_pred[i]] += 1

    def zero(self):
        self.mat.fill(0)
    
    def ratings(self):
        return np.array([self.lista, self.listb])

    def errors(self):
        """
        Calculate different error types
        :return: vetors of true postives (tp) false negatives (fn), false positives (fp) and true negatives (tn)
                 pos 0 is first class, pos 1 is second class etc.
        """
        tp = np.asarray(np.diag(self.mat).flatten(),dtype='float')
        fn = np.asarray(np.sum(self.mat, axis=1).flatten(),dtype='float') - tp
        fp = np.asarray(np.sum(self.mat, axis=0).flatten(),dtype='float') - tp
        tn = np.asarray(np.sum(self.mat)*np.ones(self.n_classes).flatten(),dtype='float') - tp - fn - fp
        return tp, fn, fp, tn

    def accuracy(self):
        tp, _, _, _ = self.errors()
        n_samples = np.sum(self.mat)
        return np.sum(tp) / n_samples

    def sensitivity(self):
        tp, tn, fp, fn = self.errors()
        res = tp / (tp + fn)
        res = res[~np.isnan(res)]
        return res

    def specificity(self):
        tp, tn, fp, fn = self.errors()
        res = tn / (tn + fp)
        res = res[~np.isnan(res)]
        return res

    def ppv(self):
        tp, tn, fp, fn = self.errors()
        res = tp / (tp + fp)
        res = res[~np.isnan(res)]
        return res

    def npv(self):
        tp, tn, fp, fn = self.errors()
        res = tn / (tn + fn)
        res = res[~np.isnan(res)]
        return res

    def fpr(self):
        tp, tn, fp, fn = self.errors()
        res = fp / (fp + tn)
        res = res[~np.isnan(res)]
        return res

    def fdr(self):
        tp, tn, fp, fn = self.errors()
        res = fp / (tp + fp)
        res = res[~np.isnan(res)]
        return res

    def f1(self):
        tp, tn, fp, fn = self.errors()
        res = (2*tp) / (2*tp + fp + fn)
        res = res[~np.isnan(res)]
        return res

    def prevalence(self):
        tp, tn, fp, fn = self.errors()
        res = (tp + fn) / (tp + fn + fp + tn)
        res = res[~np.isnan(res)]
        return res

    def matthews_correlation(self):
        tp, tn, fp, fn = self.errors()
        numerator = tp*tn - fp*fn
        denominator = np.sqrt((tp + fp)*(tp + fn)*(tn + fp)*(tn + fn))
        res = numerator / denominator
        res = res[~np.isnan(res)]
        return res

    def matrix(self):
        return self.mat

    def binary_pretty_print(self, class_labels=['0','1']): 
        '''
        Improves the built-in sklearn confusion matrix with associated values
        C: np.ndarray, shape (2, 2) as given by sklearn's confusion_matrix
        '''

        from sklearn.metrics import confusion_matrix
        import matplotlib.pyplot as plt
        import numpy as np
        
        C = confusion_matrix(self.lista, self.listb)

        assert C.shape == (2,2), "Confusion matrix should be from binary classification only."
        
        tn = C[0,0]; fp = C[0,1]; fn = C[1,0]; tp = C[1,1];

        NP = fn+tp
        NN = tn+fp
        N  = NP+NN

        fig = plt.figure(figsize=(8,8))
        ax  = fig.add_subplot(111)
        ax.imshow(C, interpolation='nearest', cmap=plt.cm.gray)

        # Draw the grid boxes
        ax.set_xlim(-0.5,2.5)
        ax.set_ylim(2.5,-0.5)
        ax.plot([-0.5,2.5],[0.5,0.5], '-k', lw=2)
        ax.plot([-0.5,2.5],[1.5,1.5], '-k', lw=2)
        ax.plot([0.5,0.5],[-0.5,2.5], '-k', lw=2)
        ax.plot([1.5,1.5],[-0.5,2.5], '-k', lw=2)

        # Set xlabels
        ax.set_xlabel('Predicted Label', fontsize=16)
        ax.set_xticks([0,1,2])
        ax.set_xticklabels(class_labels + [''])
        ax.xaxis.set_label_position('top')
        ax.xaxis.tick_top()
        ax.xaxis.set_label_coords(0.34,1.06)

        # Set ylabels
        ax.set_ylabel('True Label', fontsize=16, rotation=90)
        ax.set_yticklabels(class_labels + [''],rotation=90)
        ax.set_yticks([0,1,2])
        ax.yaxis.set_label_coords(-0.09,0.65)


        ax.text(0,0,
                'True Neg: %d\n(Num Neg: %d)'%(tn,NN),
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))

        ax.text(0,1,
                'False Neg: %d'%fn,
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))

        ax.text(1,0,
                'False Pos: %d'%fp,
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))


        ax.text(1,1,
                'True Pos: %d\n(Num Pos: %d)'%(tp,NP),
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))

        ax.text(2,0,
                'False Pos Rate: %.2f'%(fp / (fp+tn+0.)),
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))

        ax.text(2,1,
                'True Pos Rate: %.2f'%(tp / (tp+fn+0.)),
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))

        ax.text(2,2,
                'Accuracy: %.2f'%((tp+tn+0.)/N),
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))

        ax.text(0,2,
                'Neg Pre Val: %.2f'%(1-fn/(fn+tn+0.)),
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))

        ax.text(1,2,
                'Pos Pred Val: %.2f'%(tp/(tp+fp+0.)),
                va='center',
                ha='center',
                bbox=dict(fc='w',boxstyle='round,pad=1'))


        plt.tight_layout()
        return plt